package List;
# List module: linked lists using references, OO version with stringification
#
#	build the equivalent of the Haskell-style data decln:
#		data list = nil or cons( scalar head, list tail )
#	representing nil as [], and cons(h,t) as [h,t] (a 2-tuple)

use strict;
use warnings;
use Function::Parameters qw(:strict);
use Data::Dumper;


our $as_string_limit = 0;        # limit for as_string ("display first N").  0 means unlimited.


# $l = List->nil - return an empty list
#
fun nil( $class )
{
	return bless [], $class;
}

# $l = List->cons( $head, $tail ) - return a new list node.
#	$head becomes the head of the new list, and $tail the tail.
#
fun cons( $class, $head, $tail )
{
	return bless [ $head, $tail ], $class;
}

# $isnil = $list->isnil - return true iff the given list is nil
#
method isnil()
{
	die "List::isnil, bad list ".Dumper($self) unless
		$self->isa("List") && (@$self == 0 || @$self == 2);
	return @$self == 0 ? 1 : 0;
}

# ( $head, $tail ) = $list->headtail - break nonempty list into head and tail
#
method headtail()
{
	die "List::headtail, bad list ".Dumper($self) unless
		$self->isa("List") && (@$self == 0 || @$self == 2);
	die "List::headtail, empty list\n" if @$self == 0;
	my( $h, $t ) = @$self;
	return ( $h, $t );
}

# $len = $list->len - return the length of the given list
#
method len()
{
	my $len = 0;
	while( ! $self->isnil )
	{
		( my $h, $self ) = $self->headtail;
		$len++;
	}
	return $len;
}

# $list = $l1->append( $l2 ) - build a new list comprising all of
#			       list $l1 followed by all of list $l2
#
method append( $l2 )
{
	return $l2 if $self->isnil;
	my( $h, $t ) = $self->headtail;
	return List->cons( $h, $t->append($l2) );
}

# $revlist = $list->rev - return the reverse of $list
#
method rev()
{
	my $result = List->nil;
	while( ! $self->isnil )
	{
		( my $h, $self ) = $self->headtail;
		$result = List->cons( $h, $result );
	}
	return $result;
}

# $str = $list->as_string( [$limit] ) - return the printable form of the given list
#
method as_string( $limit = $as_string_limit )
{
	my $str = "";
	for( my $i = 1; ! $self->isnil && ($limit == 0 || $i <= $limit); $i++ )
	{
		( my $h, $self ) = $self->headtail;
		$str .= "$h,";
	}
	chop $str;                            # remove trailing ','
	$str .= "..." unless $self->isnil;    # must show that list has been cutoff!
	return "[ $str ]";
}

# Operator overloading of "stringify" (turn into a string)
use overload '""' => \&overload_as_string;
fun overload_as_string( $list, $x, $y )    # don’t care about last 2 params
{
	return $list->as_string;
}

1;
